"read.FCS" <- function(fileName, myFCSobj.name="", fcs.type=NULL,                        
                       ##    fcs.version = c("FCS1.0","FCS2.0", "FCS3.0"),
                       ## the fcs.version is read from the header

                       fcs.byte.size=2,
                       fcs.signed=TRUE,
                       fcs.endian = "auto", # c("auto","big","little"),
                       use.FCS.shortnames = FALSE,
                       no.names=FALSE,
                       UseS3=FALSE,
                       MY.DEBUG=TRUE) {

  ## This new version of read.FCS can now read FCS3.0 files
  ## will check only the range of FCS3.0 files and
  ## range and size of all other FCS versions
  
  
  ## This function takes a fileName and optional version number, and
  ## creates a dataset for analysis, from a flow-cytometry experiment.

  ########################################################################
  # input variables    description
  ########################################################################
  # fileName           file location as a string
  # fix.datarange      boolean, switch the endian by using endian="swap"
  #                    to range the data from 0 to 1024
  # FCSversion         optional version number
  # MY.DEBUG           boolean
  ########################################################################
  # output description:
  #    creates a FCS file with the following components
  #
  #      $data         dataframe with a working dataset
  #      $metadata
  ########################################################################

  ## Thanks to Peter Rabinovitch for information and sample data.

  ## get the params from the signature
  ## fcs.version <- ifelse(is.null(fcs.type),
  ##                     match.arg(fcs.version),fcs.type$version)
  fcs.byte.size <- ifelse(is.null(fcs.type),
                          fcs.byte.size,fcs.type$byte.size)
  fcs.signed <- ifelse(is.null(fcs.type),
                       fcs.signed,fcs.type$signed)
  fcs.endian <- ifelse(is.null(fcs.type),
                       fcs.endian,fcs.type$endian)
  ## Uses binary streams to read the data

  ## 4 parts:
  ## - HEADER,
  ## - TEXT,
  ## - DATA,
  ## - ANALYSIS (optional)

  ## open binary connection for reading
  zz <- file(description=fileName,
             open="rb")

  #### Get Header

  ## numbers imply bytes.  This might not work everywhere.
  seek(zz,0)
  my.metadata.Header <- readChar(zz,58)
  my.metadata.indices <- list(version = substr(my.metadata.Header,1,6),
                             TextStart = as.integer(substr(my.metadata.Header,10,18)),
                             TextEnd = as.integer(substr(my.metadata.Header,19,26)),
                             TextLength = (as.integer(substr(my.metadata.Header,19,26)) -
                                           as.integer(substr(my.metadata.Header,10,18)) + 1),
                             DataStart = as.integer(substr(my.metadata.Header,27,34)),
                             DataEnd = as.integer(substr(my.metadata.Header,35,42)),
                             DataLength = (as.integer(substr(my.metadata.Header,35,42)) -
                                           as.integer(substr(my.metadata.Header,27,34)) + 1),
                             AnalysisStart = as.integer(substr(my.metadata.Header,43,50)),
                             AnalysisEnd = as.integer(substr(my.metadata.Header,51,58)))

  if (MY.DEBUG) {
    cat("Start/Stop:")
    if (length(my.metadata.indices)==9){
      print(my.metadata.indices)
    } else {
      warning("read.FCS: my.metadata.indices is not the correct length of 9")
    }
  }
  
  fcs.version <- my.metadata.indices$version
  ## Read the text section:
  seek(zz,my.metadata.indices[["TextStart"]]) 
  my.metadata.Text <- readChar(zz,my.metadata.indices[["TextLength"]])

  ##  my.metadata.delimiter <- "\f"
  my.metadata.delimiter <- substr(my.metadata.Text,0,1)

  ## removing all end and beginning spaces
  my.metadata.Text.verbose <- unlist(strsplit(my.metadata.Text, split=""))
  pos.space<- which(my.metadata.Text.verbose==" ")
  first <- 1
  last  <- length(my.metadata.Text.verbose)
  continue.end <- continue.begin <- TRUE

  if (my.metadata.Text.verbose[last] != my.metadata.delimiter){
    ## There is an error in the metadata
    ## There might be some more metadata, but it will be truncated
    last.pos.delim <- max(which(my.metadata.Text.verbose == my.metadata.delimiter))
    warning(paste("Metadata Length Discrepancy: TextLength =",
                  my.metadata.indices[["TextLength"]],
                  "but valid metadata length =", last.pos.delim, sep=" "))
    warning("The metadata will be truncated up to the last delimiter.")

   
    my.metadata.Text.verbose <- my.metadata.Text.verbose[1:last.pos.delim]
    last <- last.pos.delim
  }
  
  while(length(pos.space)>=1 & last>=1 &
        continue.begin==TRUE ||
        continue.end==TRUE) {
    if (max(pos.space)==last) {
      pos.space <- pos.space[pos.space!=max(pos.space)]
      last <- last-1
    }
    continue.end <- my.metadata.Text.verbose[last] != my.metadata.delimiter
    if (length(pos.space)>0 & last>0) {
      if(min(pos.space)==first){
        pos.space <- pos.space[pos.space!=min(pos.space)]
        first <- first+1
      }
    }
    continue.begin <- my.metadata.Text.verbose[first] != my.metadata.delimiter
  }

  if (my.metadata.delimiter %in% c("\\")) {
    my.metadata.delimiter <- "\\\\"
  }
  if (first>last & MY.DEBUG==TRUE) {
    print("There is no metadata.")
    my.metadata.Text.list <- NULL
  } else {
    ## There is metadata...
    my.metadata.Text.verbose<- my.metadata.Text.verbose[first:last]
    my.metadata.Text <- paste(my.metadata.Text.verbose, collapse="")
  
    ## This is a list of strings for the parameters
    my.metadata.Text.list <-unlist(strsplit(substr(my.metadata.Text,
                                                   1,length(my.metadata.Text.verbose)),
                                            my.metadata.delimiter))
 
    ##  The following should read in all the metadata that is in the file ##

    if (MY.DEBUG==TRUE) print("Obtaining the metadata.")
    my.metadata.matrix <- matrix(my.metadata.Text.list[my.metadata.Text.list!=""],
                                 ncol=2, byrow=TRUE)
    my.metadata.names <- my.metadata.matrix[,1]
    metadata.list <- as.list(my.metadata.matrix[,2])
    names(metadata.list) <- my.metadata.names
  }

  ##  Two functions
 
  param.dependent.keywords <- function(prefix,suffix) {
    ## returns a vector of strings
    ## the vector has a length of $PAR (the number of parameters)
    ## each string includes the position in the vector
    paste(prefix,1:as.integer(get.metadata("$PAR")),suffix,sep="")
  }

  get.metadata <- function(paramString) {
    ## matches the metadata name, paramString
    ## returns the value associated with the metadata name, paramString
    data.par <- ifelse(is.na(charmatch(c(paramString),
                                       my.metadata.Text.list)),
                       NA,
                       my.metadata.Text.list[charmatch(c(paramString),
                                                       my.metadata.Text.list) + 1])
    return(data.par)
  }

  ## The following are some commonly used FCS metadata parameters
  ## as specified by the FCS version

  ## FCS 2.0 

  ## JYW: not needed b/c currently all information in the header is in
  ## the metadata
  
  ## FCS2.params <- as.vector(c("$PAR",
  ##                            "$TOT",
  ##                            "$MODE"))
  
  ## FCS 3.0:  The above plus the below are REQUIRED for FCS 3.0
  ## FCS3.params.segments <- as.vector(c("$BEGINANALYSIS", "$BEGINDATA",
  ##                           "$BYTEORD", "$BEGINSTEXT", "$DATATYPE",
  ##                          "$ENDANALYSIS", "$ENDDATA", "$ENDSTEXT",
  ##                         "$NEXTDATA"))

  ## $PnB Number of bits reserved for parameter number n.
  ## 
  ## $PnE Amplification type for parameter n.
  ##
  ## $PnR Range for parameter number n.

  ## JYW: Currently everything in the header is included in the metadata
  ## FCS3.params.PARdep <-
  ## c(param.dependent.keywords("$P","B"),
  ##  param.dependent.keywords("$P","E"),
  ## param.dependent.keywords("$P","R"))

  ## FCS3.params.opt <- c("$ABRT", 
  ##                  "$BTIM", "$CELLS", "$COM", "$COMP", "$CSMODE",
  ##                "$CSVBITS", "$CYT", "$CYTSN", "$DATE", "$ETIM",
  ##              "$EXP", "$FIL", "$GATE", "$GATING", "$INST",
  ##            "$LOST", "$OP", "$PROJ", "$SMNO", "$SRC",
  ##          "$SYS", "$TIMESTEP", "$TR", "$UNICODE")
  
  ## FCS3.params.opt.PARdep <-
  ## c(param.dependent.keywords("$CSV","FLAG"),
  ##  param.dependent.keywords("$G","E"),
  ##  param.dependent.keywords("$G","F"),
  ##  param.dependent.keywords("$G","N"),
  ##  param.dependent.keywords("$G","P"),
  ##  param.dependent.keywords("$G","R"),
  ##  param.dependent.keywords("$G","S"),
  ##  param.dependent.keywords("$G","T"),
  ##  param.dependent.keywords("$G","V"),
  ##  param.dependent.keywords("$Pk",""),
  ##  param.dependent.keywords("$PKN",""),
  ##  param.dependent.keywords("$P","F"),
  ##  param.dependent.keywords("$P","G"),
  ##  param.dependent.keywords("$P","L"),
  ##  param.dependent.keywords("$P","N"),
  ##  param.dependent.keywords("$P","O"),
  ##  param.dependent.keywords("$P","P"),
  ##  param.dependent.keywords("$P","S"),
  ## param.dependent.keywords("$P","T"),
  ## param.dependent.keywords("$P","V"),
  ## param.dependent.keywords("$R","I"),
  ## param.dependent.keywords("$R","W"))
  
  ## Optional:
  ##   3.2.19 The optional FCS TEXT segment keywordsare as follows:
  ## 
  ## $ABRT Events lost due to data acquisitionelectronic coincidence.
  ## 
  ## $BTIM Clock time at beginning of data acquisition.
  ## 
  ## $CELLS Description of objects measured.
  ## 
  ## $COM Comment.
  ## 
  ## $COMP Fluorescence compensation matrix.
  ## 
  ## $CSMODE Cell subset mode, number of subsetsto which an object may belong.
  ## 
  ## $CSVBITS Number of bits used to encodea cell subset identifier.
  ## 
  ## $CSVnFLAG The bit set as a flag for subsetn.
  ## 
  ## $CYT Type of flow cytometer.
  ## 
  ## $CYTSN Flow cytometer serial number.
  ## 
  ## $DATE Date of data set acquisition.
  ## 
  ## $ETIM Clock time at end of data acquisition.
  ## 
  ## $EXP Name of investigator initiating theexperiment.
  ## 
  ## $FIL Name of the data file containingthe data set.
  ## 
  ## $GATE Number of gating parameters.
  ## 
  ## $GATING Specifies region combinations usedfor gating.
  ## 
  ## $GnE Amplification type for gating parameternumber n.
  ## 
  ## $GnF Optical filter used for gating parameternumber n.
  ## 
  ## $GnN Name of gating parameter number n.
  ## 
  ## $GnP Percent of emitted light collectedby gating parameter n.
  ## 
  ## $GnR Range of gating parameter n.
  ## 
  ## $GnS Name used for gating parameter n.
  ## 
  ## $GnT Detector type for gating parametern.
  ## 
  ## $GnV Detector voltage for gating parametern.
  ## 
  ## $INST Institution at which data acquired.
  ## 
  ## $LOST Number of events lost due to computerbusy.
  ## 
  ## $OP Name of flow cytometry operator.
  ## 
  ## $Pkn Peak channel number of univariatehistogram for parameter n.
  ## 
  ## $PKNn Count in peak channel of univariatehistogram for parameter n.
  ## 
  ## $PnF Name of optical filter for parametern.
  ## 
  ## $PnG Amplifier gain used for acquisitionof parameter n.
  ## 
  ## $PnL Excitation wavelength for parametern.
  ## 
  ## $PnN Short name for parameter n.
  ## 
  ## $PnO Excitation power for parameter n.
  ## 
  ## $PnP Percent of emitted light collectedby parameter n.
  ## 
  ## $PnS Name used for parameter n.
  ## 
  ## $PnT Detector type for parameter n.
  ## 
  ## $PnV Detector voltage for parameter n.
  ## 
  ## $PROJ Name of the experiment project.
  ## 
  ## $RnI Gating region for parameter numbern.
  ## 
  ## $RnW Window settings for gating region n.
  ## 
  ## $SMNO Specimen (tube or well) label.
  ## 
  ## $SRC Source of the specimen (patient name,cell types)
  ## 
  ## $SYS Type of computer and its operatingsystem.
  ## 
  ## $TIMESTEP Time step for time parameter.
  ## 
  ## $TR Trigger parameter and its threshold.
  ## 
  ## $UNICODE UNICODE code page for string typekeyword values.
  
  ## Use $BYTEORD to compute this, based on current endian-ness of the
  ## system (i.e. whether to swap or not)  
  ## the following is just the default, from Mac files to Intel platforms.

  ## Fixing the problem stated above, we define "big" endian to be
  ## 4321 byte order ($BYTEORD) and we define "little" endian to be
  ## 1234 and compare it to the .Platform$endian, ie, the endian of
  ## the processor in use We swap if the endian of the FCS file does
  ## not match with the endian of the current platform.
  
  if (is.na(metadata.list$"$BYTEORD")) {
    stop("There is no $BYTEORD variable in the metadata.")
  }

  byte.string <- as.vector(unlist(strsplit(metadata.list$"$BYTEORD", "")))
  position.last <- which(byte.string=="4")
  num <- 4
  while (length(position.last)==0 & num > 0) {
    num <- num -1
    position.last <- which(byte.string==as(num, "character"))
  }
  
  position1 <- which(byte.string=="1")

  if (position.last==position1){
    stop("$BYTEORD should be 4,3,2,1 or 1,2,3,4 or 1,2 or 2,1 format. ")
  }

 
  file.endian <- ifelse(position.last < position1, "big", "little")
  if (fcs.endian == "auto") {
    my.data.endian <- ifelse(.Platform$endian==file.endian,
                             .Platform$endian, "swap")
  } else {
    if (fcs.endian == "big" || fcs.endian == "little") {
      my.data.endian <- fcs.endian
    }
  }

  if (MY.DEBUG==TRUE) {
    print(paste("$BYTEORD", metadata.list$"$BYTEORD", sep=" "))
    print(paste("file endian is", file.endian, sep=" "))
    print(paste("platform endian is", .Platform$endian, sep=" "))
    print(paste("FCS object endian is", my.data.endian, sep=" "))
  }
 
  ## Get Data
  if (MY.DEBUG==TRUE){
    print("Reading in data")
  }
  seek(zz,my.metadata.indices[["DataStart"]], origin="start")

  max.byte <- as(byte.string[position.last], "numeric")
  if (max.byte==4){
    n.cell.data <-  my.metadata.indices[["DataLength"]]
  } else {
    
    ## not sure if this will work with different
    ## byte orders (ie, 1,2,3??)
    n.cell.data <- floor(my.metadata.indices[["DataLength"]]/max.byte)
  }
  cell.data <- readBin(zz,integer(),
                       signed=fcs.signed,
                       n=n.cell.data,
                       size=fcs.byte.size,
                       endian=my.data.endian)

  ## do not know why, but the data includes four extra numbers being read in
  ## at the end, which corresponds to the number 12336

  ## last.pos is the total number of observations noted in the metadata
  last.pos <- as.numeric(metadata.list$"$TOT") * as.numeric(metadata.list$"$PAR")
  if (MY.DEBUG==TRUE) {
    print("Checking the size and range of data")
  }

  metadata.ranges <- as.numeric(unlist(metadata.list[names(metadata.list) %in% param.dependent.keywords("$P", "R")]))
  metadata.min <- 0
  metadata.max <- max(metadata.ranges)
 
  checked <- FALSE
  read.data.again <- FALSE
  
  if ((last.pos < length(cell.data)) & fcs.version=="FCS3.0") {
    ## only in FCS3.0 files there are extra elements read in
    ## only subset the first elements indicated by the metadata.
    cell.data.red <- cell.data[1:last.pos]
    within.range <- (min(cell.data.red) >= metadata.min) & (max(cell.data.red) <= metadata.max)
    if (within.range==TRUE){
      if (MY.DEBUG==TRUE){
        print(paste("WARNING: Size of data [", length(cell.data),
                    "] is greater than the size noted in metadata [",
                    last.pos,"].", sep=""))
        print("WARNING: The following extra elements of the data are omitted when constructing the data.frame:")
        print(cell.data[(last.pos+1):length(cell.data)])
      }
      cell.data <- cell.data.red
      checked <- TRUE
      read.data.again <- FALSE
      mismatch.length <- FALSE
    }
  }
  
  if (checked==FALSE) {
    within.range <- (min(cell.data) >= metadata.min) & (max(cell.data) <= metadata.max)
    mismatch.length <- last.pos!=length(cell.data)
    if ((within.range==FALSE) || (mismatch.length==TRUE)){
      read.data.again <- TRUE
      if (MY.DEBUG==TRUE){
        print(paste("   Within metadata range?", within.range, sep=" "))
        print(paste("   Size of data [", length(cell.data),
                    "]; Size noted in metadata [", last.pos,
                    "]", sep=""))
      }
    } else {
      if ((within.range==TRUE) & (mismatch.length==FALSE)) {
        read.data.again <- FALSE
        if (MY.DEBUG==TRUE) {
          print("Data passes range and size check")
        }
      }
    }
  }

  ## JYW: If there is a size mismatch or a range discrepancy between the
  ## data and the metadata, we will try to fix it here
  ## when we are reading in that data
  ## rather than potentially finding out later, which could be bad!
  ## We try to read in the data again, because it might be a fcs.byte.size
  ## or a fcs.signed misspecification
  
  ## try a different byte size
  if (read.data.again==TRUE) {

    possible.signed <- c(1, 0) ## 1==TRUE, 0==FALSE
    possible.bytesize <- c(1,2,4,8)
    possible.bytesize.signed <- as.list(c(paste(possible.bytesize, possible.signed[1],
                                                sep=" "),
                                          paste(possible.bytesize, possible.signed[2],
                                                sep=" ")))

    i <- 1
    cell.data.a <- cell.data
    within.range.a <- within.range
    removed.data <- NULL
    mismatch.length.a <- mismatch.length
    
    while ((i <= length(possible.bytesize.signed)) & ((mismatch.length.a == TRUE) || (within.range.a==FALSE))) {
      possibility <- as.numeric(unlist(strsplit(possible.bytesize.signed[[i]],
                                                split=" ")))
      BSIZE <- possibility[1]
      if (possibility[2]==1){
        SIGNED <- TRUE
      } else {
        if (possibility[2]==0) {
          SIGNED <- FALSE
        }
      }
      seek(zz,my.metadata.indices[["DataStart"]], origin="start")
      cell.data.a <- readBin(zz,integer(),
                           signed=SIGNED,
                           n=n.cell.data,
                           size=BSIZE,
                           endian=my.data.endian)
      
      if (last.pos!= length(cell.data.a)){
        if ( (fcs.version=="FCS3.0") & (last.pos <length(cell.data.a))){
          ## we force the size to be equal to the metadata
          ## for FCS3.0 files we are only checking the range
          
          removed.data <- cell.data.a[(last.pos+1):length(cell.data.a)]
          cell.data.a <- cell.data.a[1:last.pos]
          mismatch.length.a <- FALSE
        } else if ((fcs.version!="FCS3.0") || (last.pos > length(cell.data.a)) ){
          mismatch.length.a <- TRUE
        }
      } else {
        if (last.pos==length(cell.data.a)) {
          mismatch.length.a <- FALSE
        }
      }
      
      within.range.a <- ((min(cell.data.a) >= metadata.min) &
                         (max(cell.data.a) <= metadata.max))
      i <- i+1
    }

    if (fcs.version=="FCS3.0"){
      failure.on.rereads <- (within.range.a==FALSE)
    } else {
      if (fcs.version!="FCS3.0") {
        failure.on.rereads <- (mismatch.length.a==TRUE) || (within.range.a==FALSE)
      }
    }
    
    if(failure.on.rereads==TRUE) {
      cat("Error: Exhausted all fcs.byte.size (1,2,4,8) and fcs.signed (TRUE, FALSE)", "\n",
          "possibilities for reading in data.", "\n",
          "There is a discrepancy between the metadata and the data.  ", "\n",
          "Situation of the first data read with default read.FCS arguments:", "\n",
                 "   :Range match=", within.range, "\n",
                 "   :size of data=", length(cell.data), "\n",
                 "   :size noted in metadata=", last.pos, "\n\n")
      stop("Error: Check the input FCS binary file for metadata and data for discrepancies.")
    } else {
      if (failure.on.rereads==FALSE) {
        cell.data <- cell.data.a
        if (MY.DEBUG==TRUE) {
          print("Re-read the data again; passed range and size check on these parameters:")
          print(paste("   fcs.byte.size, fcs.signed, fcs.endian: ", BSIZE, ",", SIGNED, ",", fcs.endian, sep=""))
          if (fcs.version=="FCS3.0") {
            print("   NOTE: This file is FCS version 3.0 and only the range NOT size has been checked.")
            
            print(paste("WARNING: Size of data [", length(cell.data) +length(removed.data),
                        "] is greater than the size noted in metadata [",
                        last.pos,"].", sep=""))
            print("WARNING: The following extra elements of the data are omitted when constructing the data.frame:")
            print(removed.data)
          }
        }
      }
    }
  }
  ## check the ranges... if there are negatives then try a different fcs.signed
  
  if (MY.DEBUG==TRUE) {
    print(paste("Constructing data.frame from N= ",length(cell.data)))
  }
  cell.df <- as.data.frame(t(matrix(cell.data,
                                    ## JYW: There appears to be a mismatch between
                                    ## what the metadata.list says is the total number of
                                    ## data points = 50,000 ($TOT=10,000 x $PAR=5)
                                    ## and what is read in from my.metadata.indices
                                    ## DataLength=200,000
                                    ncol=as.integer(metadata.list$"$TOT"))))

  if (MY.DEBUG==TRUE) print("Get names")

  if (!no.names) {
    ## If we don't want to automatically use names, don't do this.
    ## (usually, we will!)
    if (use.FCS.shortnames) {
      fcs.varnames <- sapply(param.dependent.keywords("$P","N"),
                             get.metadata)
      if (sum(is.na(fcs.varnames))==length(fcs.varnames)){
        cat("NOTE: Short names \$PnN are missing", "\n")
        cat("Long names \$PnS are used instead.", "\n")
        fcs.varnames <- sapply(param.dependent.keywords("$P","S"),
                               get.metadata)
      }
    } else {
      fcs.varnames <- sapply(param.dependent.keywords("$P","S"),
                             get.metadata)
       if (sum(is.na(fcs.varnames))==length(fcs.varnames)){
         cat("NOTE: Long names \$PnS are missing.", "\n")
         cat("Short names \$PnN are assigned to the dataset instead.", "\n")
         fcs.varnames <- sapply(param.dependent.keywords("$P","N"),
                                                 get.metadata)
       }
    }
   
    if (MY.DEBUG==TRUE) {
      print(paste("Assigning ",length(fcs.varnames),
                  " names to dataframe dims",
                  dim(cell.df)[1]," x ",dim(cell.df)[2]))
    }
    colnames(cell.df) <- fcs.varnames
  }
  
  ##
  ## Get Analysis Section
  ##

  ## normally, we do not care much about the analysis section.  In
  ## particular, we are more interested in the data and other
  ## components section.  The analysis section contains histograms and
  ## similar from FCS programs.

  ## we currently don't know what to do, and probably don't care.
  my.file.has.Analysis.Section <-  FALSE
  if (my.file.has.Analysis.Section == TRUE) {
    ## do stuff
  }

  if (MY.DEBUG==TRUE) print("Constructing return value")

  ## close the connection before exiting 
  close(zz)

  if (UseS3) {
    ## S3 version
    return(structure(list(data=cell.df,metadata=metadata.list),
                     class="FCS"))
  } else {
    ## when we get S4 classes running, then finish off by:
    return(new("FCS",
               data=as.matrix(cell.df),
               metadata=new("FCSmetadata",mode=metadata.list[["$MODE"]],
                 size=as.numeric(metadata.list[["$TOT"]]),
                 nparam=as.numeric(metadata.list[["$PAR"]]),
                 shortnames= unlist(get.metadata(param.dependent.keywords("$P", "N"))),
                 longnames=unlist(get.metadata(param.dependent.keywords("$P", "S"))),
                 paramranges=as.numeric(unlist(get.metadata(param.dependent.keywords("$P", "R")))),
                 filename=fileName,
                 objectname=myFCSobj.name,original=TRUE,
                 fcsinfo=metadata.list[!(names(metadata.list) %in% c("$MODE", "$TOT", "$PAR",
                   param.dependent.keywords("$P", "N"),
                   param.dependent.keywords("$P", "S"),
                   param.dependent.keywords("$P", "R") ))])))
  }
}

